import base64
import mimetypes
import os
from pathlib import Path
from typing import Optional

from email.mime.audio import MIMEAudio
from email.mime.base import MIMEBase
from email.mime.image import MIMEImage
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from googleapiclient import errors


from . import LibraryContext, keyword


def get_size_format(b, factor=1024, suffix="B"):
    """
    Scale bytes to its proper byte format
    e.g:
        1253656 => '1.20MB'
        1253656678 => '1.17GB'
    """
    for unit in ["", "K", "M", "G", "T", "P", "E", "Z"]:
        if b < factor:
            return f"{b:.2f}{unit}{suffix}"
        b /= factor
    return f"{b:.2f}Y{suffix}"


def clean(text):
    # clean text for creating a folder
    return "".join(c if c.isalnum() else "_" for c in text)


class GmailKeywords(LibraryContext):
    """Class for Google Gmail API

    **Note:** The Gmail API does not work with _service accounts_

    For more information about Google Gmail API link_.

    .. _link: https://developers.google.com/gmail/api
    """

    def __init__(self, ctx):
        super().__init__(ctx)
        self.service = None

    @keyword(tags=["init", "gmail"])
    def init_gmail(
        self,
        service_account: str = None,
        credentials: str = None,
        use_robocorp_vault: Optional[bool] = None,
        scopes: list = None,
        token_file: str = None,
    ) -> None:
        """Initialize Google Gmail client

        :param service_account: file path to service account file
        :param credentials: file path to credentials file
        :param use_robocorp_vault: use credentials in `Robocorp Vault`
        :param scopes: list of extra authentication scopes
        :param token_file: file path to token file
        """
        gmail_scopes = ["gmail.send", "gmail.compose", "gmail.modify", "gmail.labels"]
        if scopes:
            gmail_scopes += scopes
        self.service = self.init_service(
            service_name="gmail",
            api_version="v1",
            scopes=gmail_scopes,
            service_account_file=service_account,
            credentials_file=credentials,
            use_robocorp_vault=use_robocorp_vault,
            token_file=token_file,
        )

    def create_message(
        self,
        to: str,
        subject: str,
        message_text: str,
        attachments: list = None,
    ):
        """Create a message for an email.

        :param to: message recipient
        :param subject: message subject
        :param message_text: message body text
        :param attachment: list of files to add as message attachments
        :return: An object containing a base64url encoded email object
        """
        mimeMessage = MIMEMultipart()
        mimeMessage["to"] = to
        mimeMessage["subject"] = subject
        mimeMessage.attach(MIMEText(message_text, "plain"))

        for at in attachments:
            self.add_attachment_to_message(mimeMessage, at)
        return {"raw": base64.urlsafe_b64encode(mimeMessage.as_bytes()).decode()}

    def add_attachment_to_message(self, mimeMessage, attachment):
        content_type, encoding = mimetypes.guess_type(attachment)

        if content_type is None or encoding is not None:
            content_type = "application/octet-stream"
        main_type, sub_type = content_type.split("/", 1)
        if main_type == "text":
            with open(attachment, "r") as fp:
                msg = MIMEText(fp.read(), _subtype=sub_type)
        elif main_type == "image":
            with open(attachment, "rb") as fp:
                msg = MIMEImage(fp.read(), _subtype=sub_type)
        elif main_type == "audio":
            with open(attachment, "rb") as fp:
                msg = MIMEAudio(fp.read(), _subtype=sub_type)
        else:
            with open(attachment, "rb") as fp:
                msg = MIMEBase(main_type, sub_type)
                msg.set_payload(fp.read())
        filename = os.path.basename(attachment)
        msg.add_header("Content-Disposition", "attachment", filename=filename)
        mimeMessage.attach(msg)

    @keyword(tags=["gmail"])
    def send_message(
        self,
        sender: str,
        to: str,
        subject: str,
        message_text: str,
        attachments: list = None,
    ):
        """Send an email message.

        :param sender: message sender
        :param to: message recipient
        :param subject: message subject
        :param message_text: message body text
        :param attachment: list of files to add as message attachments
        :return: sent message

        Example:

        .. code-block:: robotframework

            ${attachments}=  Create List
            ...  ${CURDIR}${/}random.txt
            ...  ${CURDIR}${/}source.png
            Send Message    me
            ...    mika@robocorp.com
            ...    message subject
            ...    body of the message
            ...    ${attachments}
        """
        if not self.service:
            raise AssertionError("Gmail service has not been initialized")
        attachments = attachments or []
        message = self.create_message(to, subject, message_text, attachments)
        try:
            response = (
                self.service.users()
                .messages()
                .send(userId=sender, body=message)
                .execute()
            )
            self.logger.debug("Message Id: %s" % response["id"])
        except errors.HttpError as he:
            self.logger.warning(str(he))
            raise he
        return response

    def set_list_parameters(self, user_id, query, label_ids, max_results, include_spam):
        parameters = {"userId": user_id, "q": query}
        if label_ids:
            parameters["labelIds"] = ",".join(label_ids)
        if max_results:
            parameters["maxResults"] = max_results
        if include_spam:
            parameters["includeSpamTrash"] = include_spam
        return parameters

    def set_headers_to_message_dict(self, payload, message_id, response):
        headers = payload.get("headers")
        message_dict = {"id": message_id, "label_ids": response["labelIds"]}
        for header in headers:
            name = header.get("name")
            value = header.get("value")
            if name.lower() == "from":
                message_dict["from"] = value
            if name.lower() == "to":
                message_dict["to"] = value
            if name.lower() == "subject":
                message_dict["subject"] = value
            if name.lower() == "date":
                message_dict["date"] = value
        return message_dict

    def handle_mimetypes(self, parsed_parts, part, msg, folder_name):
        filename = part.get("filename")
        mimetype = part.get("mimeType")
        body = part.get("body")
        data = body.get("data")
        filesize = body.get("size")
        part_headers = part.get("headers")
        if mimetype == "text/plain":
            if data:
                text = base64.urlsafe_b64decode(data).decode()
                parsed_parts.append({"text/plain": text})
        elif mimetype == "text/html":
            if not filename:
                filename = "message.html"
            filepath = os.path.join(folder_name, filename)

            with open(filepath, "wb") as f:
                data = base64.urlsafe_b64decode(data)
                parsed_parts.append({"text/html": data, "path": filepath})
                f.write(data)
        else:
            for part_header in part_headers:
                part_header_name = part_header.get("name")
                part_header_value = part_header.get("value")
                if part_header_name == "Content-Disposition":
                    if "attachment" in part_header_value:
                        # we get the attachment ID
                        # and make another request to get the attachment itself
                        self.logger.info(
                            "Saving the file: %s, size:%s"
                            % (filename, get_size_format(filesize))
                        )
                        attachment_id = body.get("attachmentId")
                        attachment = (
                            self.service.users()
                            .messages()
                            .attachments()
                            .get(
                                id=attachment_id,
                                userId="me",
                                messageId=msg["id"],
                            )
                            .execute()
                        )
                        data = attachment.get("data")
                        filepath = os.path.join(folder_name, filename)
                        if data:
                            parsed_parts.append(
                                {"attachment": filepath, "id": attachment_id}
                            )
                            with open(filepath, "wb") as f:
                                f.write(base64.urlsafe_b64decode(data))

    @keyword(tags=["gmail"])
    def list_messages(
        self,
        user_id: str,
        query: str,
        folder_name: str = None,
        label_ids: list = None,
        max_results: int = None,
        include_json: bool = False,
        include_spam: bool = False,
    ):
        """List messages

        :param user_id: user's email address. The special value me can
         be used to indicate the authenticated user.
        :param query: message query
        :param folder_name: path where attachments are saved, default current
         directory
        :param label_ids: message label ids
        :param max_results: maximum number of message to return
        :param include_json: include original response json
        :param include_spam: include messages from SPAM and TRASH
        :return: messages

        Example:

        .. code-block:: robotframework

            ${messages}=    List Messages    me
            ...    from:mika@robocorp.com
            ...    folder_name=${CURDIR}${/}target
            ...    include_json=True
            FOR    ${msg}    IN    @{messages}
                Log Many    ${msg}
            END
        """
        parameters = self.set_list_parameters(
            user_id, query, label_ids, max_results, include_spam
        )
        folder_name = Path(folder_name) if folder_name else Path().absolute()
        messages = []
        try:
            response = self.service.users().messages().list(**parameters).execute()
            message_ids = [
                m["id"] for m in response["messages"] if "messages" in response.keys()
            ]
            for message_id in message_ids:
                response = (
                    self.service.users()
                    .messages()
                    .get(userId=user_id, id=message_id)
                    .execute()
                )
                payload = response["payload"]
                message_dict = self.set_headers_to_message_dict(
                    payload, message_id, response
                )
                if include_json:
                    message_dict["json"] = response
                parts = payload.get("parts")
                parsed_parts = self.parse_parts(
                    message_id, response, parts, folder_name
                )
                message_dict["parts"] = parsed_parts
                messages.append(message_dict)
        except errors.HttpError as he:
            self.logger.warning(str(he))
            raise he
        return messages

    def parse_parts(self, msg_id, msg, parts, folder_name):
        """
        Utility function that parses the content of an email partition
        """
        if msg_id and msg_id not in str(folder_name):
            try:
                folder_name = folder_name / msg_id
                folder_name.mkdir(parents=True, exist_ok=True)
            except FileExistsError:
                pass

        parsed_parts = []
        if parts:
            for part in parts:
                if part.get("parts"):
                    # recursively call this function when we see that a part
                    # has parts inside
                    self.parse_parts(None, msg, part.get("parts"), folder_name)
                self.handle_mimetypes(parsed_parts, part, msg, folder_name)

        return parsed_parts



from typing import Dict, Optional

from google.cloud import videointelligence

from . import LibraryContext, keyword, to_feature


class VideoIntelligenceKeywords(LibraryContext):
    """Keywords for Google Video Intelligence API"""

    def __init__(self, ctx):
        super().__init__(ctx)
        self.service = None

    @keyword(tags=["init", "video intelligence"])
    def init_video_intelligence(
        self,
        service_account: str = None,
        use_robocorp_vault: Optional[bool] = None,
        token_file: str = None,
    ) -> None:
        """Initialize Google Cloud Video Intelligence client

        :param service_account: file path to service account file
        :param use_robocorp_vault: use credentials in `Robocorp Vault`
        :param token_file: file path to token file
        """
        self.service = self.init_service_with_object(
            videointelligence.VideoIntelligenceServiceClient,
            service_account,
            use_robocorp_vault,
            token_file,
        )

    @keyword(tags=["video intelligence"])
    def annotate_video(
        self,
        video_file: str = None,
        video_uri: str = None,
        features: str = None,
        output_uri: str = None,
        json_file: str = None,
        timeout: int = 300,
    ) -> Dict:
        """Annotate video

        Possible values for features:

        - FEATURE_UNSPECIFIED, Unspecified.
        - LABEL_DETECTION, Label detection. Detect objects, such as dog or flower.
        - SHOT_CHANGE_DETECTION, Shot change detection.
        - EXPLICIT_CONTENT_DETECTION, Explicit content detection.
        - SPEECH_TRANSCRIPTION, Speech transcription.
        - TEXT_DETECTION, OCR text detection and tracking.
        - OBJECT_TRACKING, Object detection and tracking.
        - LOGO_RECOGNITION, Logo detection, tracking, and recognition.

        If `video_uri` is given then that is used even if `video_file` is given.

        :param video_file: local file path to input video
        :param video_uri: Google Cloud Storage URI to input video
        :param features: list of annotation features to detect,
            defaults to LABEL_DETECTION,SHOT_CHANGE_DETECTION
        :param output_uri: Google Cloud Storage URI to store response json
        :param json_file: json target to save result
        :param timeout: timeout for operation in seconds
        :return: annotate result

        **Examples**

        **Robot Framework**

        .. code-block:: robotframework

            ${result}=   Annotate Video   video_uri=gs://videointelligence/movie.mp4
            ...  features=TEXT_DETECTION,LABEL_DETECTION
            ...  output_uri=gs://videointelligence/movie_annotations.json
            ...  json_file=${CURDIR}${/}videoannotations.json
        """
        if features is None:
            features_in = [
                videointelligence.Feature.LABEL_DETECTION,
                videointelligence.Feature.SHOT_CHANGE_DETECTION,
            ]
        else:
            features_in = [to_feature(feature) for feature in features.split(",")]
        parameters = {"features": features_in}
        if video_uri:
            parameters["input_uri"] = video_uri
        elif video_file:
            video_context = videointelligence.VideoContext()
            with open(video_file, "rb") as file:
                input_content = file.read()
                parameters["input_content"] = input_content
                parameters["video_context"] = video_context
        if output_uri:
            parameters["output_uri"] = output_uri

        operation = self.service.annotate_video(request=parameters)
        result = operation.result(timeout=timeout)
        self.write_json(json_file, result)
        return result



from enum import Enum
from google.cloud import language_v1, videointelligence


class TextType(Enum):
    """Possible text types."""

    TEXT = language_v1.Document.Type.PLAIN_TEXT
    HTML = language_v1.Document.Type.HTML


class UpdateAction(Enum):
    """Possible file update actions."""

    trash = 1
    untrash = 2
    star = 3
    unstar = 4


def to_texttype(value):
    """Convert value to TextType enum."""
    if isinstance(value, TextType):
        return int(value.value)

    sanitized = str(value).upper().strip().replace(" ", "_")
    try:
        return int(TextType[sanitized].value)
    except KeyError as err:
        raise ValueError(f"Unknown text type: {value}") from err


class VideoFeature(Enum):
    """Possible video features."""

    FEATURE_UNSPECIFIED = videointelligence.Feature.FEATURE_UNSPECIFIED
    LABEL_DETECTION = videointelligence.Feature.LABEL_DETECTION
    SHOT_CHANGE_DETECTION = videointelligence.Feature.SHOT_CHANGE_DETECTION
    EXPLICIT_CONTENT_DETECTION = videointelligence.Feature.EXPLICIT_CONTENT_DETECTION
    SPEECH_TRANSCRIPTION = videointelligence.Feature.SPEECH_TRANSCRIPTION
    TEXT_DETECTION = videointelligence.Feature.TEXT_DETECTION
    OBJECT_TRACKING = videointelligence.Feature.OBJECT_TRACKING
    LOGO_RECOGNITION = videointelligence.Feature.LOGO_RECOGNITION


def to_feature(value):
    """Convert value to VideoFeature enum."""
    if isinstance(value, VideoFeature):
        return int(value.value)

    sanitized = str(value).upper().strip().replace(" ", "_")
    try:
        return int(VideoFeature[sanitized].value)
    except KeyError as err:
        raise ValueError(f"Unknown video feature: {value}") from err



from typing import Dict, Optional
from google.cloud import language_v1


from . import LibraryContext, keyword, TextType, to_texttype


class NaturalLanguageKeywords(LibraryContext):
    """Keywords for Google Cloud Natural Language API"""

    def __init__(self, ctx):
        super().__init__(ctx)
        self.service = None

    @keyword(tags=["init", "natural language"])
    def init_natural_language(
        self,
        service_account: str = None,
        use_robocorp_vault: Optional[bool] = None,
        token_file: str = None,
    ) -> None:
        """Initialize Google Cloud Natural Language client

        :param service_account: file path to service account file
        :param use_robocorp_vault: use credentials in `Robocorp Vault`
        :param token_file: file path to token file
        """
        self.service = self.init_service_with_object(
            language_v1.LanguageServiceClient,
            service_account,
            use_robocorp_vault,
            token_file,
        )

    @keyword(tags=["natural language"])
    def analyze_sentiment(
        self,
        text: str = None,
        text_file: str = None,
        file_type: TextType = TextType.TEXT,
        json_file: str = None,
        lang: str = None,
    ) -> Dict:
        """Analyze sentiment in a text file

        :param text: source text
        :param text_file: source text file
        :param file_type: type of text, PLAIN_TEXT (default) or HTML
        :param json_file: json target to save result, defaults to None
        :param lang: language code of the source, defaults to None
        :return: analysis response

        # For list of supported languages:
        # https://cloud.google.com/natural-language/docs/languages

        **Examples**

        **Robot Framework**

        .. code-block:: robotframework

            ${result}=   Analyze Sentiment  ${text}
            ${result}=   Analyze Sentiment  text_file=${CURDIR}${/}test.txt
        """
        return self._analyze_handler(
            text, text_file, file_type, json_file, lang, "sentiment"
        )

    @keyword(tags=["natural language"])
    def classify_text(
        self,
        text: str = None,
        text_file: str = None,
        file_type: TextType = TextType.TEXT,
        json_file: str = None,
        lang: str = None,
    ) -> Dict:
        """Classify text

        :param text: source text
        :param text_file: source text file
        :param file_type: type of text, PLAIN_TEXT (default) or HTML
        :param json_file: json target to save result, defaults to None
        :param lang: language code of the source, defaults to None
        :return: classify response

        # For list of supported languages:
        # https://cloud.google.com/natural-language/docs/languages

        **Examples**

        **Robot Framework**

        .. code-block:: robotframework

            ${result}=   Classify Text  ${text}
            ${result}=   Classify Text  text_file=${CURDIR}${/}test.txt
        """
        return self._analyze_handler(
            text, text_file, file_type, json_file, lang, "classify"
        )

    def _analyze_handler(
        self, text, text_file, file_type, json_file, lang, analyze_method
    ):
        file_type = to_texttype(file_type)
        parameters = {"type_": file_type}
        if text:
            parameters["content"] = text
        elif text_file:
            with open(text_file, "r") as f:
                parameters["content"] = f.read()
        else:
            raise AttributeError("Either 'text' or 'text_file' must be given")

        if lang is not None:
            parameters["language"] = lang

        document = language_v1.Document(**parameters)
        if analyze_method == "classify":
            response = self.service.classify_text(document=document)
        elif analyze_method == "sentiment":
            # Available values: NONE, UTF8, UTF16, UTF32
            # encoding_type = enums.EncodingType.UTF8
            response = self.service.analyze_sentiment(
                document=document, encoding_type="UTF8"
            )
        self.write_json(json_file, response)
        return response



from typing import Optional

from . import (
    LibraryContext,
    keyword,
)


class AppsScriptKeywords(LibraryContext):
    """Class for Google Apps Script API

    For more information about Google Apps Script API link_.

    .. _link: https://developers.google.com/apps-script/api
    """

    def __init__(self, ctx):
        super().__init__(ctx)
        self.service = None

    @keyword(tags=["init", "apps script"])
    def init_apps_script(
        self,
        service_account: str = None,
        credentials: str = None,
        use_robocorp_vault: Optional[bool] = None,
        scopes: list = None,
        token_file: str = None,
    ) -> None:
        """Initialize Google Apps Script client

        :param service_account: file path to service account file
        :param credentials: file path to credentials file
        :param use_robocorp_vault: use credentials in `Robocorp Vault`
        :param scopes: list of extra authentication scopes
        :param token_file: file path to token file
        """
        apps_scopes = ["script.projects", "drive.scripts", "script.external_request"]
        if scopes:
            apps_scopes += scopes
        self.service = self.init_service(
            service_name="script",
            api_version="v1",
            scopes=apps_scopes,
            service_account_file=service_account,
            credentials_file=credentials,
            use_robocorp_vault=use_robocorp_vault,
            token_file=token_file,
        )

    @keyword(tags=["apps script"])
    def run_script(
        self, script_id: str, function_name: str, parameters: dict = None
    ) -> None:
        """Run the Google Apps Script function

        :param script_id: Google Script identifier
        :param function_name: name of the script function
        :param parameters: script function parameters as a dictionary
        :raises AssertionError: thrown when Google Script returns errors

        Example:

        .. code-block:: robotframework

            &{params}=    Create Dictionary  formid=aaad4232  formvalues=1,2,3
            ${response}=  Run Script    abc21397283712da  submit_form   ${params}
            Log Many   ${response}
        """
        request = {
            "function": function_name,
        }
        if parameters:
            request["parameters"] = [parameters]
        response = (
            self.service.scripts()
            .run(
                body=request,
                scriptId=script_id,
            )
            .execute()
        )
        if "error" in response.keys():
            raise AssertionError(response["error"])
        return response
